import re

import schwifty
from banal import ensure_list

from ftm_analyze.analysis.util import TAG_COUNTRY, TAG_EMAIL, TAG_IBAN, TAG_PHONE

EMAIL_REGEX = re.compile(r"[A-Z0-9._%+-]+@[A-Z0-9.-]+\.[A-Z]{2,}", re.IGNORECASE)
PHONE_REGEX = re.compile(r"(\+?[\d\-\(\)\/\s]{5,}\d{2})", re.IGNORECASE)
IBAN_REGEX = re.compile(
    r"\b([a-zA-Z]{2} ?[0-9]{2} ?[a-zA-Z0-9]{4} ?[0-9]{7} ?([a-zA-Z0-9]?){0,16})\b",
    re.IGNORECASE,
)

REGEX_TYPES = {
    EMAIL_REGEX: TAG_EMAIL,
    PHONE_REGEX: TAG_PHONE,
    IBAN_REGEX: TAG_IBAN,
}


def check_iban(value: str) -> str | None:
    try:
        return schwifty.IBAN(value, allow_invalid=True).is_valid
    except schwifty.exceptions.SchwiftyException:
        return


def extract_patterns(entity, text):
    for pattern, prop in REGEX_TYPES.items():
        for match in pattern.finditer(text):
            match_text = match.group(0)
            value = prop.type.clean(match_text, proxy=entity)
            if value is None:
                continue
            if prop == TAG_IBAN:
                if not check_iban(value):
                    continue
            yield (prop, value)
            for country in ensure_list(prop.type.country_hint(value)):
                yield (TAG_COUNTRY, country)



"""
Annotate entities for use with
https://www.elastic.co/docs/reference/elasticsearch/plugins/mapper-annotated-text-usage
"""

from functools import cache
from typing import Iterable, Self
from urllib.parse import quote_plus as qs

from anystore.types import StrGenerator
from followthemoney import E, EntityProxy, Property, Schema, model, registry
from followthemoney.compare import _normalize_names
from normality import collapse_spaces
from pydantic import BaseModel

from ftm_analyze.analysis.util import (
    TAG_COMPANY,
    TAG_EMAIL,
    TAG_IBAN,
    TAG_LOCATION,
    TAG_NAME,
    TAG_PERSON,
    TAG_PHONE,
)

ANNOTATED = "__annotated__"
MENTION_PROPS = {
    TAG_NAME.name,
    TAG_PERSON.name,
    TAG_COMPANY.name,
    TAG_EMAIL.name,
    TAG_PHONE.name,
    TAG_IBAN.name,
    TAG_LOCATION.name,
}
PER = "Person"
ORG = "Organization"
LEG = "LegalEntity"
NAMED = {TAG_COMPANY.name, TAG_PERSON.name, TAG_NAME.name}
SKIP_CHARS = "()[]"


def entity_fingerprints(entity: EntityProxy) -> set[str]:
    """Get the set of entity name fingerprints"""
    # FIXME
    return set(_normalize_names(entity.schema, entity.names))


def make_fingerprints(schemata: set[Schema], names: Iterable[str]) -> set[str]:
    """Mimic `fingerprints.generate`"""
    # FIXME
    fps: set[str] = set()
    for schema in schemata:
        fps.update(set(_normalize_names(schema, names)))
    return fps


def clean_text(text: str) -> str:
    """Clean the text before annotation: Remove [...](...) patterns"""
    for c in SKIP_CHARS:
        text = text.replace(c, " ")
    return collapse_spaces(text) or ""


@cache
def get_extend(schema: Schema) -> set[str]:
    schemata: set[str] = {schema.name}
    for s in schema.extends:
        if s.name not in ("Thing", "Value"):
            schemata.update(get_extend(s))
    return schemata


class Annotation(BaseModel):
    """lorem ipsum [Mrs. Jane Doe](f_doe+jane&s_Person&s_LegalEntity&p_namesMentioned&p_peopleMentioned) dolor sit amet"""  # noqa: B950

    value: str
    names: set[str] = set()
    schemata: set[str] = set()
    props: set[str] = set()

    @property
    def is_name(self) -> bool:
        return bool(NAMED & self.props)

    @property
    def fingerprints(self) -> set[str]:
        schemata = self._schemata or {model[LEG]}
        if self.is_name:
            return make_fingerprints(schemata, self._names)
        return set()

    @property
    def _names(self) -> set[str]:
        if self.is_name:
            return set([self.value, *self.names])
        return set()

    @property
    def _props(self) -> set[str]:
        if self.is_name:
            return set([*self.props, TAG_NAME.name])
        return self.props

    @property
    def _schemata(self) -> set[Schema]:
        if self.is_name:
            schemata = {model.get(s) for s in self.schemata}
            schemata = {s for s in schemata if s}
            if not schemata:
                if TAG_PERSON.name in self.props:
                    return {model[s] for s in get_extend(model[PER])}
                if TAG_COMPANY.name in self.props:
                    return {model[s] for s in get_extend(model[ORG])}
            return schemata
        return set()

    def get_query(self) -> str:
        parts: set[str] = set()
        for fp in self.fingerprints:
            parts.add(f"f_{qs(fp)}")
        for prop in self._props:
            parts.add(f"p_{prop}")
        for schema in self._schemata:
            parts.add(f"s_{schema.name}")
        return "&".join(sorted(parts))

    @property
    def repl(self) -> str | None:
        query = self.get_query()
        if query:
            return f"[{self.value}]({query})"

    def annotate(self, text: str) -> str:
        repl = self.repl
        if repl:
            return text.replace(self.value, repl)
        return text

    def update(self, a: Self) -> None:
        if self.value != a.value:
            raise ValueError(f"Invalid value from update annotation: `{a.value}`")
        self.names.update(a.names)
        self.schemata.update(a.schemata)
        self.props.update(a.props)

    @classmethod
    def from_entity(cls, value: str, e: EntityProxy) -> Self:
        if not e.schema.is_a("LegalEntity"):
            raise ValueError(f"Invalid schema: `{e.schema}` (not a LegalEntity)")
        props = {TAG_NAME.name}
        if e.schema.is_a(ORG):
            props.add(TAG_COMPANY.name)
        if e.schema.is_a(PER):
            props.add(TAG_PERSON.name)
        return cls(
            value=value,
            names=set(e.names),
            schemata=get_extend(e.schema),
            props=props,
        )


class Annotator:
    def __init__(self, entity: EntityProxy) -> None:
        self.entity = entity
        self.annotations: dict[str, Annotation] = {}

    def add(self, a: Annotation) -> None:
        if not a.props & MENTION_PROPS:
            # skip non mentions
            return
        if a.value in self.annotations:
            self.annotations[a.value].update(a)
        else:
            self.annotations[a.value] = a

    def add_tag(self, prop: Property | str, value: str) -> None:
        if isinstance(prop, Property):
            prop = prop.name
        a = Annotation(props={prop}, value=value)
        self.add(a)

    def add_mention(self, value: str, e: EntityProxy) -> None:
        a = Annotation.from_entity(value, e)
        self.add(a)

    def annotate_text(self, text: str) -> str:
        for a in self.annotations.values():
            text = a.annotate(text)
        return text

    def get_texts(self) -> StrGenerator:
        for text in self.entity.get_type_values(registry.text):
            text = clean_text(text)
            annotated = self.annotate_text(text)
            if annotated:
                yield annotated


def annotate_entity(e: E) -> E:
    if not e.schema.is_a("Analyzable"):
        return e
    annotator = Annotator(e)
    schema = model["Analyzable"]
    for prop in schema.properties:
        for value in e.get(prop):
            annotator.add_tag(prop, value)
    for text in annotator.get_texts():
        e.add("indexText", f"{ANNOTATED} {text}")
    return e



from typing import Optional

import typer
from anystore.cli import ErrorHandler
from anystore.logging import configure_logging, get_logger
from ftmq.io import smart_read_proxies, smart_write_proxies
from rich.console import Console
from spacy.cli.download import download
from typing_extensions import Annotated

from ftm_analyze import __version__, logic
from ftm_analyze.settings import Settings

settings = Settings()
cli = typer.Typer(
    no_args_is_help=True,
    pretty_exceptions_enable=True,
    pretty_exceptions_short=not settings.debug,
)
console = Console(stderr=True)

log = get_logger(__name__)

IN = Annotated[
    str, typer.Option(..., "-i", help="Input entities uri (file, http, s3...)")
]
OUT = Annotated[
    str, typer.Option(..., "-o", help="Output entities uri (file, http, s3...)")
]


@cli.callback(invoke_without_command=True)
def cli_base(
    version: Annotated[Optional[bool], typer.Option(..., help="Show version")] = False,
    settings: Annotated[
        Optional[bool], typer.Option(..., help="Show current settings")
    ] = False,
):
    if version:
        print(__version__)
        raise typer.Exit()
    if settings:
        print(Settings())
        raise typer.Exit()
    configure_logging()


@cli.command("download-spacy")
def cli_download():
    """
    Download required spacy models based on current settings
    """
    console.print(settings.ner_models)
    models = settings.ner_models.model_dump()
    for model in models.values():
        download(model)


@cli.command("analyze")
def cli_analyze(
    in_uri: IN = "-",
    out_uri: OUT = "-",
    resolve_mentions: Annotated[
        bool, typer.Option(help="Resolve known mentions via `juditha`")
    ] = True,
    annotate: Annotated[
        bool, typer.Option(help="Annotate extracted patterns, names and mentions")
    ] = True,
):
    """
    Analyze a stream of entities.
    """
    with ErrorHandler(log):
        entities = smart_read_proxies(in_uri)
        results = logic.analyze_entities(entities, resolve_mentions, annotate)
        smart_write_proxies(out_uri, results)



import logging

from banal import ensure_list
from countrytagger import tag_place

log = logging.getLogger(__name__)


def location_country(location) -> list[str]:
    res = tag_place(location)
    if res is None:
        return []
    return ensure_list(res[2])



from collections import defaultdict

from anystore.logging import get_logger

from ftm_analyze.analysis.ft_type_model import FTTypeModel
from ftm_analyze.settings import Settings

log = get_logger(__name__)
settings = Settings()


class TagAggregatorFasttext(object):
    def __init__(
        self,
        model_path=settings.ner_type_model_path,
        confidence: float | None = settings.ner_type_model_confidence,
    ):
        self.values = defaultdict(set)
        self.model = FTTypeModel(str(model_path))
        self.confidence = confidence

    def add(self, prop, value):
        if value is None:
            return
        key = prop.type.node_id_safe(value)
        self.values[(key, prop)].add(value)

    def results(self):
        for (key, prop), values in self.values.items():
            values.discard(None)
            if not values:
                continue
            values = list(values)
            labels, confidences = self.model.confidence(values)
            if not self.confidence:
                # very messy
                yield (key, prop, values)
                continue
            for label, confidence in zip(labels, confidences):
                if label == "trash" or (
                    self.confidence and confidence < self.confidence
                ):
                    break
            else:
                yield (key, prop, values)

    def __len__(self):
        return len(self.values)


class TagAggregator(object):
    MAX_TAGS = 10000

    def __init__(self):
        self.values = defaultdict(list)

    def add(self, prop, value):
        key = prop.type.node_id_safe(value)
        if key is None:
            return

        if (key, prop) not in self.values:
            if len(self.values) > self.MAX_TAGS:
                return

        self.values[(key, prop)].append(value)

    def results(self):
        for (key, prop), values in self.values.items():
            yield (key, prop, values)

    def __len__(self):
        return len(self.values)



from typing import Generator, Iterable

from anystore.decorators import error_handler
from anystore.io import logged_items
from anystore.logging import get_logger
from followthemoney.proxy import EntityProxy

from ftm_analyze.analysis.analyzer import Analyzer
from ftm_analyze.settings import Settings

log = get_logger(__name__)
settings = Settings()


@error_handler(logger=log)
def analyze_entity(
    entity: EntityProxy,
    resolve_mentions: bool | None = settings.resolve_mentions,
    annotate: bool | None = settings.annotate,
) -> Generator[EntityProxy, None, None]:
    """
    Analyze an Entity.

    Args:
        entity: The entity proxy
        resolve_mentions: Convert known mentions into its actual entities via
            `juditha`
        annotate: Annotate extracted patterns, names and mentions in `indexText`

    Yields:
        A generator of entity fragments
    """
    analyzer = Analyzer(entity, resolve_mentions, annotate)
    analyzer.feed(entity)
    yield from analyzer.flush()


def analyze_entities(
    entities: Iterable[EntityProxy],
    resolve_mentions: bool | None = settings.resolve_mentions,
    annotate: bool | None = settings.annotate,
) -> Generator[EntityProxy, None, None]:
    for e in logged_items(entities, "Analyze", item_name="Entity", logger=log):
        yield from analyze_entity(e, resolve_mentions, annotate)



from functools import lru_cache
from typing import Generator

import spacy
from anystore.logging import get_logger
from fingerprints import clean_entity_prefix
from followthemoney import Property
from followthemoney.types import registry
from normality import collapse_spaces
from rigour.langs import list_to_alpha3

from ftm_analyze.analysis.country import location_country
from ftm_analyze.analysis.util import TAG_COMPANY, TAG_COUNTRY, TAG_LOCATION, TAG_PERSON
from ftm_analyze.settings import Settings

log = get_logger(__name__)
settings = Settings()

NAME_MAX_LENGTH = 100
NAME_MIN_LENGTH = 4
# https://spacy.io/api/annotation#named-entities
SPACY_TYPES = {
    "PER": TAG_PERSON,
    "PERSON": TAG_PERSON,
    "ORG": TAG_COMPANY,
    "LOC": TAG_LOCATION,
    "GPE": TAG_LOCATION,
}
NER_MODELS = settings.ner_models.model_dump()


def clean_name(text):
    if text is None or len(text) > NAME_MAX_LENGTH:
        return
    text = clean_entity_prefix(text)
    text = collapse_spaces(text)
    if text is None or len(text) <= NAME_MIN_LENGTH or " " not in text:
        return
    return text


@lru_cache(maxsize=5)
def _load_model(model):
    """Load the spaCy model for the specified language"""
    return spacy.load(model)


def get_models(entity):
    """Iterate over the NER models applicable to the given entity."""
    languages = entity.get_type_values(registry.language)
    models = set()
    for lang in list_to_alpha3(languages):
        model = NER_MODELS.get(lang)
        if model is not None:
            models.add(model)
    for model in models:
        yield _load_model(model)


def extract_entities(entity, text) -> Generator[tuple[Property, str], None, None]:
    for model in get_models(entity):
        # log.debug("NER tagging %d chars (%s)", len(text), model.lang)
        doc = model(text)
        for ent in doc.ents:
            prop = SPACY_TYPES.get(ent.label_)
            if prop is None:
                continue
            if prop in (TAG_COMPANY, TAG_PERSON, TAG_LOCATION):
                name = clean_name(ent.text)
                if name:
                    yield prop, name
            if prop == TAG_LOCATION:
                for country in location_country(ent.text):
                    yield TAG_COUNTRY, country



from pathlib import Path

from anystore.settings import BaseSettings
from pydantic_settings import BaseSettings as _BaseSettings
from pydantic_settings import SettingsConfigDict


class NerModels(_BaseSettings):
    """
    Easily overwrite specific language model for specific languages via:

    `FTM_ANALYZE_NER_MODELS_DEU=de_core_news_lg`
    """

    eng: str = "en_core_web_sm"
    deu: str = "de_core_news_sm"
    fra: str = "fr_core_news_sm"
    spa: str = "es_core_news_sm"
    rus: str = "ru_core_news_sm"
    por: str = "pt_core_news_sm"
    ron: str = "ro_core_news_sm"
    mkd: str = "mk_core_news_sm"
    ell: str = "el_core_news_sm"
    pol: str = "pl_core_news_sm"
    ita: str = "it_core_news_sm"
    lit: str = "lt_core_news_sm"
    nld: str = "nl_core_news_sm"
    nob: str = "nb_core_news_sm"
    nor: str = "nb_core_news_sm"
    dan: str = "da_core_news_sm"


class Settings(BaseSettings):
    """
    `ftm-analyze` settings management using
    [pydantic-settings](https://docs.pydantic.dev/latest/concepts/pydantic_settings/)

    Note:
        All settings can be set via environment variables or `.env` file,
        prepending `FTM_ANALYZE_` (except for those with another alias)
    """

    model_config = SettingsConfigDict(
        env_prefix="ftm_analyze_",
        env_nested_delimiter="_",
        env_file=".env",
        extra="ignore",
    )

    ner_type_model_path: Path = Path("./models/model_type_prediction.ftz")
    """Local path to ftm type predict model"""

    ner_type_model_confidence: float = 0.85
    """Minimum confidence for ftm type predict model"""

    lid_model_path: Path = Path("./models/lid.176.ftz")
    """Local path to lid model"""

    ner_models: NerModels = NerModels()
    """Spacy models"""

    resolve_mentions: bool = True
    """Resolve known mentions via `juditha`"""

    annotate: bool = True
    """Insert annotations into `indexText` for resolved mentions"""



import logging

import fasttext
import numpy as np
from normality import normalize

log = logging.getLogger(__name__)


class SingletonDecorator:
    def __init__(self, klass):
        self.klass = klass
        self.instance = None

    def __call__(self, *args, **kwds):
        if self.instance is None:
            self.instance = self.klass(*args, **kwds)
        return self.instance


@SingletonDecorator
class FTTypeModel(object):
    def __init__(self, model_path: str):
        self.model_path = model_path
        self.model = fasttext.load_model(model_path)
        self.n_labels = len(self.model.get_labels())
        self.max_entropy = -1 * np.log(1.0 / self.n_labels)

    def confidence(self, values):
        labels, scores = self.model.predict(self._clean_input(values), k=-1)
        # confidence = 1 - (entropy / max entropy) ∈ [0, 1]
        confidence = 1 + (scores * np.log(scores)).sum(axis=1) / self.max_entropy
        return list(self._clean_labels(labels)), confidence

    def _clean_input(self, values):
        return [normalize(v, lowercase=True, latinize=True) for v in values]

    def _clean_labels(self, labels):
        for label in labels:
            yield label[0].replace("__label__", "")



from followthemoney import model
from normality import collapse_spaces

ANALYZABLE = model["Analyzable"]
DOCUMENT = model["Document"]
TAG_NAME = ANALYZABLE.properties["namesMentioned"]
TAG_PERSON = ANALYZABLE.properties["peopleMentioned"]
TAG_COMPANY = ANALYZABLE.properties["companiesMentioned"]
TAG_LANGUAGE = ANALYZABLE.properties["detectedLanguage"]
TAG_COUNTRY = ANALYZABLE.properties["detectedCountry"]
TAG_EMAIL = ANALYZABLE.properties["emailMentioned"]
TAG_PHONE = ANALYZABLE.properties["phoneMentioned"]
TAG_IBAN = ANALYZABLE.properties["ibanMentioned"]
TAG_LOCATION = ANALYZABLE.properties["locationMentioned"]


def text_chunks(texts, sep=" ", max_chunk=25000):
    """Pre-chew text snippets for NLP and pattern matching."""
    for text in texts:
        text = collapse_spaces(text)
        if text is None or len(text) < 5:
            continue
        # Crudest text splitting code in documented human history.
        # Most of the time, a single page of text is going to be
        # 3000-4000 characters, so this really only kicks in if
        # something weird is happening in the first place.
        for idx in range(0, len(text), max_chunk):
            yield text[idx : idx + max_chunk]



from followthemoney.proxy import EntityProxy
from openaleph_procrastinate import defer
from openaleph_procrastinate.app import make_app
from openaleph_procrastinate.model import DatasetJob
from openaleph_procrastinate.tasks import task

from ftm_analyze.logic import analyze_entities

app = make_app(__loader__.name)
ORIGIN = "analyze"


@task(app=app)
def analyze(job: DatasetJob) -> None:
    entities: list[EntityProxy] = list(job.load_entities())
    with job.get_writer() as bulk:
        for entity in analyze_entities(entities):
            bulk.put(entity, origin=ORIGIN)
            entities.append(entity)
    defer.index(app, job.dataset, entities, **job.context)



import logging
from itertools import chain
from typing import Generator

import juditha
from followthemoney import model
from followthemoney.proxy import EntityProxy
from followthemoney.types import registry
from followthemoney.util import make_entity_id
from rigour.names import pick_name

from ftm_analyze.analysis.aggregate import TagAggregator, TagAggregatorFasttext
from ftm_analyze.analysis.extract import extract_entities
from ftm_analyze.analysis.language import detect_languages
from ftm_analyze.analysis.patterns import extract_patterns
from ftm_analyze.analysis.util import (
    ANALYZABLE,
    DOCUMENT,
    TAG_COMPANY,
    TAG_IBAN,
    TAG_PERSON,
    text_chunks,
)
from ftm_analyze.annotate import ANNOTATED, NAMED, Annotator
from ftm_analyze.settings import Settings

log = logging.getLogger(__name__)
settings = Settings()


STOP_EXTENDS = (
    "Thing",
    "Value",
)


class Analyzer(object):
    MENTIONS = {TAG_COMPANY: "Organization", TAG_PERSON: "Person"}

    def __init__(
        self,
        entity: EntityProxy,
        resolve_mentions: bool | None = settings.resolve_mentions,
        annotate: bool | None = settings.annotate,
    ):
        self.entity = model.make_entity(entity.schema)
        self.entity.id = entity.id
        self.aggregator_entities = TagAggregatorFasttext()
        self.aggregator_patterns = TagAggregator()
        self.resolve_mentions = resolve_mentions
        self.annotate = annotate
        self.annotator = Annotator(entity)

    def feed(self, entity):
        if not entity.schema.is_a(ANALYZABLE):
            return
        texts = entity.get_type_values(registry.text)
        for text in text_chunks(texts):
            detect_languages(self.entity, text)
            for prop, tag in extract_entities(self.entity, text):
                self.aggregator_entities.add(prop, tag)
            for prop, tag in extract_patterns(self.entity, text):
                self.aggregator_patterns.add(prop, tag)

    def flush(self) -> Generator[EntityProxy, None, None]:
        countries = set()
        results = list(
            chain(
                self.aggregator_entities.results(), self.aggregator_patterns.results()
            )
        )

        for key, prop, _ in results:
            if prop.type == registry.country:
                countries.add(key)

        mention_ids = set()
        entity_ids = set()

        for key, prop, values in results:
            label = values[0]
            if prop.type == registry.name:
                label = pick_name(values)

            resolved = False
            proxy = None
            if values and self.resolve_mentions and prop.name in NAMED:
                # convert mentions in actual entities if their names are known
                for value in values:
                    lookup = juditha.lookup(value)
                    if lookup is not None:
                        proxy = self.make_entity(key, values, lookup, countries)
                        entity_ids.add(proxy.id)
                        yield proxy
                        resolved = True

                        break

            elif prop == TAG_IBAN:
                for value in values:
                    iban_proxy = self.make_bankaccount(value, countries)
                    entity_ids.add(iban_proxy.id)
                    yield iban_proxy

            # annotate mentions
            if self.annotate and proxy is not None:
                for value in values:
                    self.annotator.add_mention(value, proxy)

            if not resolved:
                # otherwise create Mention entities
                schema = self.MENTIONS.get(prop)
                if schema is not None and self.entity.schema.is_a(DOCUMENT):
                    mention = self.make_mention(prop, key, values, schema, countries)
                    mention_ids.add(mention.id)
                    yield mention

                # annotate named mentions
                if self.annotate:
                    for value in values:
                        self.annotator.add_tag(prop, value)

            self.entity.add(prop, label, cleaned=True, quiet=True)

        if self.annotate:
            for text in self.annotator.get_texts():
                self.entity.add("indexText", f"{ANNOTATED} {text}")

        if len(results):
            log.debug(
                "Extracted %d prop values, %d mentions, %d entities [%s]: %s",
                len(results),
                len(mention_ids),
                len(entity_ids),
                self.entity.schema.name,
                self.entity.id,
            )

            yield self.entity

    def make_mention(
        self, prop, key, values, schema, countries: set[str] | None = None
    ) -> EntityProxy:
        mention = model.make_entity("Mention")
        mention.make_id("mention", self.entity.id, prop, key)
        mention.add("resolved", make_entity_id(key))
        mention.add("document", self.entity.id)
        mention.add("name", values)
        mention.add("detectedSchema", schema)
        mention.add("contextCountry", countries)
        return mention

    def make_entity(
        self, key, values, result, countries: set[str] | None = None
    ) -> EntityProxy:
        proxy = model.make_entity(result.common_schema)
        proxy.id = make_entity_id(key)
        proxy.add("proof", self.entity.id)
        proxy.add("name", values)
        proxy.add("name", result.caption)
        proxy.add("country", countries)
        return proxy

    def make_bankaccount(
        self, value: str, countries: set[str] | None = None
    ) -> EntityProxy:
        bank_account = model.make_entity("BankAccount")
        bank_account.make_id("iban", self.entity.id, value)
        bank_account.add("proof", self.entity.id)
        bank_account.add("iban", value)
        bank_account.add("country", countries)
        return bank_account



from functools import cache

import fasttext
from anystore.logging import get_logger

from ftm_analyze.settings import Settings

log = get_logger(__name__)
settings = Settings()

THRESHOLD = 0.6


@cache
def get_lid_model():
    return fasttext.load_model(str(settings.lid_model_path))


def detect_languages(entity, text, k=1):
    """Given a list of lines, return a list of (line, lang)"""
    if entity.has("language", quiet=True) or entity.has("detectedLanguage"):
        # Don't detect if a language is hard-coded.
        return
    entity.pop("detectedLanguage")
    langs = get_lid_model().predict(text, k=k)
    for lang, score in zip(*langs):
        if score <= THRESHOLD:
            continue
        # fasttext labels are prefixed, with '__label__' by default
        lang = lang.replace("__label__", "")
        log.debug("Detected (%s chars): %s -> %.3f", len(text), lang, score)
        entity.add("detectedLanguage", lang)
